#include <CGAL/Exact_predicates_inexact_constructions_kernel.h>
#include <CGAL/Mesh_triangulation_3.h>
#include <CGAL/Mesh_complex_3_in_triangulation_3.h>
#include <CGAL/Mesh_criteria_3.h>
#include <CGAL/Implicit_mesh_domain_3.h>
#include <CGAL/Mesh_domain_with_polyline_features_3.h>
#include <CGAL/make_mesh_3.h>

#include <vector>
#include <map>
#include <Eigen/Dense>

#define MESH_OUTPUT 0

#if MESH_OUTPUT
#include <MeshIO.hh>
#endif

// Kernel
typedef CGAL::Exact_predicates_inexact_constructions_kernel K;
using FT          = K::FT;
using Point       = K::Point_3;

using Polyline_3 = std::vector<Point>;
using Polylines  = std::vector<Polyline_3>;
using namespace CGAL::parameters;

#include "features.inl"
#include "features_perturbed.inl"

Polylines featureCurvesNone = {};

using Vec3d = Eigen::Vector3d;
struct Box {
    Box(const Vec3d &minCorner, const Vec3d &maxCorner)
        : m_minCorner(minCorner), m_maxCorner(maxCorner) { }
    double signedDistance(const Vec3d &p) const {
        // Box is union of 6 halfspaces
        auto d = (m_minCorner - p).cwiseMax(p - m_maxCorner).eval();
        return std::min(d.maxCoeff(), 0.0) + d.cwiseMax(Vec3d::Zero()).norm();
    }
private:
    Vec3d m_minCorner, m_maxCorner;
};

struct Cylinder {
    Cylinder(const Vec3d &c, const Vec3d &dir, double r)
        : center(c), axis(dir), radius(r) { axis /= axis.norm(); }

    double signedDistance(Vec3d p) const {
        p -= center;
        double perpDist = (p - axis * axis.dot(p)).norm();
        return perpDist - radius;
    }

    Vec3d center, axis;
    double radius;
};

double exp_smin(const std::vector<double> &values, double k) {
    double res = 0;
    for (double v : values)
        res += exp(-k * v);
    return -log(res) / k;
}

struct CylinderArrangement {
    std::vector<Cylinder> cylinders;
    double smoothness;

    CylinderArrangement(double circleRadius, double smoothness)
        : smoothness(smoothness)
    {
        cylinders.emplace_back(Vec3d{ 1,  0,  0}, Vec3d{-1,  1,  0}, circleRadius);
        cylinders.emplace_back(Vec3d{ 0,  1,  0}, Vec3d{-1, -1,  0}, circleRadius);
        cylinders.emplace_back(Vec3d{-1,  0,  0}, Vec3d{ 1, -1,  0}, circleRadius);
        cylinders.emplace_back(Vec3d{ 0, -1,  0}, Vec3d{ 1,  1,  0}, circleRadius);

        cylinders.emplace_back(Vec3d{ 1,  0,  0}, Vec3d{-1,  0,  1}, circleRadius);
        cylinders.emplace_back(Vec3d{ 0,  0,  1}, Vec3d{-1,  0, -1}, circleRadius);
        cylinders.emplace_back(Vec3d{-1,  0,  0}, Vec3d{ 1,  0, -1}, circleRadius);
        cylinders.emplace_back(Vec3d{ 0,  0, -1}, Vec3d{ 1,  0,  1}, circleRadius);
    }

    double signedDistance(const Vec3d &p) const {
        std::vector<double> dists;
        for (const auto &c : cylinders)
            dists.push_back(c.signedDistance(p));
        return exp_smin(dists, 1 / smoothness);
    }
};

template<class SDFunc>
struct ClippedSDFunc : public SDFunc {
    using SDFunc::SDFunc;
    FT operator()(const Point &p) const {
        static const Box clippingBox = Box({-1, -1, -1}, {1, 1, 1});
        Vec3d pp(p[0], p[1], p[2]);
        return std::max(SDFunc::signedDistance(pp),
                        clippingBox.signedDistance(pp));
    }
};

using SDFunc = ClippedSDFunc<CylinderArrangement>;

// Domain
using Mesh_domain = CGAL::Mesh_domain_with_polyline_features_3<CGAL::Implicit_mesh_domain_3<SDFunc,K>>;

// Triangulation
using Tr = typename CGAL::Mesh_triangulation_3<Mesh_domain>::type;
using C3t3 = CGAL::Mesh_complex_3_in_triangulation_3<Tr, typename Mesh_domain::Corner_index,
                                                typename Mesh_domain::Curve_segment_index>;
using Mesh_criteria = CGAL::Mesh_criteria_3<Tr>;

int main(int argc, const char *argv[]) {
    const double circleRadius = 0.25;
    // NOTE: Perturbed feature curves were extracted for smoothness = 0.020025
    const double smoothness   = 0.02;

    std::map<std::string, const Polylines *> curveCollection = {
        {"original",  &featureCurves},
        {"perturbed", &featureCurvesPerturbed},
        {"none",      &featureCurvesNone},
    };

    if ((argc != 2) || (curveCollection.count(argv[1]) == 0)) {
        std::cerr << "usage: ./self_contained original|perturbed|none" << std::endl;
        exit(-1);
    }

    const Polylines *fc = curveCollection.at(argv[1]);
    SDFunc sdf(circleRadius, smoothness);

    // bounding sphere radius to enclose [-1, 1]^3 with small margin.
    double r = sqrt(3) + 1e-2;

    Mesh_domain domain(sdf,
            K::Sphere_3(Point(CGAL::ORIGIN), r * r), 1e-4);
    domain.add_features(fc->begin(), fc->end());

    // Mesh criteria
    Mesh_criteria criteria(facet_angle            = 30.0,
                           facet_size             = 0.02,
                           facet_distance         = 1e-3,
                           cell_radius_edge_ratio = 2.0,
                           cell_size              = 0.30,
                           edge_size              = 0.015);

    C3t3 c3t3 = CGAL::make_mesh_3<C3t3>(domain, criteria);
    std::cout << "got " << c3t3.number_of_cells_in_complex()
              << " tetrahedra" << std::endl;

#if MESH_OUTPUT
    // Output geometry
    const Tr &tr = c3t3.triangulation();
    std::map<typename C3t3::Vertex_handle, size_t> V;

    std::vector<MeshIO::IOVertex > vertices;
    std::vector<MeshIO::IOElement> elements;

    size_t i = 0;
    for (auto it = tr.finite_vertices_begin(); it != tr.finite_vertices_end(); ++it) {
        V[it] = i++;
        Point p = it->point();
        vertices.emplace_back(CGAL::to_double(p.x()),
                              CGAL::to_double(p.y()),
                              CGAL::to_double(p.z()));
    }

    for (auto it = c3t3.cells_in_complex_begin(); it != c3t3.cells_in_complex_end(); ++it) {
        elements.emplace_back(V[it->vertex(0)],
                              V[it->vertex(1)],
                              V[it->vertex(2)],
                              V[it->vertex(3)]);
    }

    MeshIO::save("geometry.msh", vertices, elements);
#endif
}
